@@ -1,5 +1,5 @@
 /*******************************************************************************
-    Copyright (c) 2015-2022 NVIDIA Corporation
+    Copyright (c) 2015-2023 NVIDIA Corporation
 
     Permission is hereby granted, free of charge, to any person obtaining a copy
     of this software and associated documentation files (the "Software"), to
@@ -29,6 +29,7 @@
 #include "uvm_tracker.h"
 #include "uvm_gpu.h"
 #include "uvm_va_space_mm.h"
+#include "uvm_processors.h"
 
 static bool uvm_is_valid_vma_range(struct mm_struct *mm, NvU64 start, NvU64 length)
 {
@@ -143,14 +144,19 @@ static NV_STATUS split_span_as_needed(uvm_va_space_t *va_space,
     return split_as_needed(va_space, end_addr, split_needed_cb, data);
 }
 
-static bool preferred_location_is_split_needed(const uvm_va_policy_t *policy, void *data)
+typedef struct
 {
     uvm_processor_id_t processor_id;
+    int nid;
+} preferred_location_split_params_t;
 
-    UVM_ASSERT(data);
+static bool preferred_location_is_split_needed(const uvm_va_policy_t *policy, void *data)
+{
+    preferred_location_split_params_t *params = (preferred_location_split_params_t *)data;
 
-    processor_id = *(uvm_processor_id_t*)data;
-    return !uvm_id_equal(processor_id, policy->preferred_location);
+    UVM_ASSERT(params);
+
+    return !uvm_va_policy_preferred_location_equal(policy, params->processor_id, params->nid);
 }
 
 static NV_STATUS preferred_location_unmap_remote_pages(uvm_va_block_t *va_block,
@@ -176,7 +182,9 @@ static NV_STATUS preferred_location_unmap_remote_pages(uvm_va_block_t *va_block,
     mapped_mask = uvm_va_block_map_mask_get(va_block, preferred_location);
 
     if (uvm_processor_mask_test(&va_block->resident, preferred_location)) {
-        const uvm_page_mask_t *resident_mask = uvm_va_block_resident_mask_get(va_block, preferred_location);
+        const uvm_page_mask_t *resident_mask = uvm_va_block_resident_mask_get(va_block,
+                                                                              preferred_location,
+                                                                              NUMA_NO_NODE);
 
         if (!uvm_page_mask_andnot(&va_block_context->caller_page_mask, mapped_mask, resident_mask))
             goto done;
@@ -219,12 +227,14 @@ static NV_STATUS preferred_location_set(uvm_va_space_t *va_space,
                                         NvU64 base,
                                         NvU64 length,
                                         uvm_processor_id_t preferred_location,
+                                        int preferred_cpu_nid,
                                         uvm_va_range_t **first_va_range_to_migrate,
                                         uvm_tracker_t *out_tracker)
 {
     uvm_va_range_t *va_range, *va_range_last;
     const NvU64 last_address = base + length - 1;
     bool preferred_location_is_faultable_gpu = false;
+    preferred_location_split_params_t split_params;
     NV_STATUS status;
 
     uvm_assert_rwsem_locked_write(&va_space->lock);
@@ -236,11 +246,13 @@ static NV_STATUS preferred_location_set(uvm_va_space_t *va_space,
                                                                       preferred_location);
     }
 
+    split_params.processor_id = preferred_location;
+    split_params.nid = preferred_cpu_nid;
     status = split_span_as_needed(va_space,
                                   base,
                                   last_address + 1,
                                   preferred_location_is_split_needed,
-                                  &preferred_location);
+                                  &split_params);
     if (status != NV_OK)
         return status;
 
@@ -264,7 +276,7 @@ static NV_STATUS preferred_location_set(uvm_va_space_t *va_space,
                 return NV_ERR_INVALID_DEVICE;
         }
 
-        status = uvm_va_range_set_preferred_location(va_range, preferred_location, mm, out_tracker);
+        status = uvm_va_range_set_preferred_location(va_range, preferred_location, preferred_cpu_nid, mm, out_tracker);
         if (status != NV_OK)
             return status;
 
@@ -282,7 +294,12 @@ static NV_STATUS preferred_location_set(uvm_va_space_t *va_space,
     if (!mm)
         return NV_ERR_INVALID_ADDRESS;
 
-    return uvm_hmm_set_preferred_location(va_space, preferred_location, base, last_address, out_tracker);
+    return uvm_hmm_set_preferred_location(va_space,
+                                          preferred_location,
+                                          preferred_cpu_nid,
+                                          base,
+                                          last_address,
+                                          out_tracker);
 }
 
 NV_STATUS uvm_api_set_preferred_location(const UVM_SET_PREFERRED_LOCATION_PARAMS *params, struct file *filp)
@@ -295,6 +312,7 @@ NV_STATUS uvm_api_set_preferred_location(const UVM_SET_PREFERRED_LOCATION_PARAMS
     uvm_va_range_t *first_va_range_to_migrate = NULL;
     struct mm_struct *mm;
     uvm_processor_id_t preferred_location_id;
+    int preferred_cpu_nid = NUMA_NO_NODE;
     bool has_va_space_write_lock;
     const NvU64 start = params->requestedBase;
     const NvU64 length = params->length;
@@ -316,6 +334,15 @@ NV_STATUS uvm_api_set_preferred_location(const UVM_SET_PREFERRED_LOCATION_PARAMS
     // If the CPU is the preferred location, we don't have to find the associated uvm_gpu_t
     if (uvm_uuid_is_cpu(&params->preferredLocation)) {
         preferred_location_id = UVM_ID_CPU;
+        preferred_cpu_nid = params->preferredCpuNumaNode;
+
+        if (preferred_cpu_nid != -1 &&
+            (!nv_numa_node_has_memory(preferred_cpu_nid) ||
+             !node_isset(preferred_cpu_nid, node_possible_map) ||
+             uvm_va_space_find_gpu_with_memory_node_id(va_space, preferred_cpu_nid))) {
+            status = NV_ERR_INVALID_ARGUMENT;
+            goto done;
+        }
     }
     else {
         // Translate preferredLocation into a live GPU ID, and check that this
@@ -353,7 +380,14 @@ NV_STATUS uvm_api_set_preferred_location(const UVM_SET_PREFERRED_LOCATION_PARAMS
         goto done;
     }
 
-    status = preferred_location_set(va_space, mm, start, length, preferred_location_id, &first_va_range_to_migrate, &local_tracker);
+    status = preferred_location_set(va_space,
+                                    mm,
+                                    start,
+                                    length,
+                                    preferred_location_id,
+                                    preferred_cpu_nid,
+                                    &first_va_range_to_migrate,
+                                    &local_tracker);
     if (status != NV_OK)
         goto done;
 
@@ -427,6 +461,7 @@ NV_STATUS uvm_api_unset_preferred_location(const UVM_UNSET_PREFERRED_LOCATION_PA
                                     params->requestedBase,
                                     params->length,
                                     UVM_ID_INVALID,
+                                    NUMA_NO_NODE,
                                     NULL,
                                     &local_tracker);
 
@@ -636,9 +671,12 @@ static NV_STATUS va_block_set_read_duplication_locked(uvm_va_block_t *va_block,
 
     uvm_assert_mutex_locked(&va_block->lock);
 
+    // Force CPU page residency to be on the preferred NUMA node.
+    va_block_context->make_resident.dest_nid = uvm_va_range_get_policy(va_block->va_range)->preferred_nid;
+
     for_each_id_in_mask(src_id, &va_block->resident) {
         NV_STATUS status;
-        uvm_page_mask_t *resident_mask = uvm_va_block_resident_mask_get(va_block, src_id);
+        uvm_page_mask_t *resident_mask = uvm_va_block_resident_mask_get(va_block, src_id, NUMA_NO_NODE);
 
         // Calling uvm_va_block_make_resident_read_duplicate will break all
         // SetAccessedBy and remote mappings
@@ -685,7 +723,6 @@ static NV_STATUS va_block_unset_read_duplication_locked(uvm_va_block_t *va_block
     uvm_page_mask_t *break_read_duplication_pages = &va_block_context->caller_page_mask;
     const uvm_va_policy_t *policy = uvm_va_range_get_policy(va_block->va_range);
     uvm_processor_id_t preferred_location = policy->preferred_location;
-    uvm_processor_mask_t accessed_by = policy->accessed_by;
 
     uvm_assert_mutex_locked(&va_block->lock);
 
@@ -695,7 +732,7 @@ static NV_STATUS va_block_unset_read_duplication_locked(uvm_va_block_t *va_block
     // If preferred_location is set and has resident copies, give it preference
     if (UVM_ID_IS_VALID(preferred_location) &&
         uvm_processor_mask_test(&va_block->resident, preferred_location)) {
-        uvm_page_mask_t *resident_mask = uvm_va_block_resident_mask_get(va_block, preferred_location);
+        uvm_page_mask_t *resident_mask = uvm_va_block_resident_mask_get(va_block, preferred_location, NUMA_NO_NODE);
         bool is_mask_empty = !uvm_page_mask_and(break_read_duplication_pages,
                                                 &va_block->read_duplicated_pages,
                                                 resident_mask);
@@ -723,7 +760,7 @@ static NV_STATUS va_block_unset_read_duplication_locked(uvm_va_block_t *va_block
         if (uvm_id_equal(processor_id, preferred_location))
             continue;
 
-        resident_mask = uvm_va_block_resident_mask_get(va_block, processor_id);
+        resident_mask = uvm_va_block_resident_mask_get(va_block, processor_id, NUMA_NO_NODE);
         is_mask_empty = !uvm_page_mask_and(break_read_duplication_pages,
                                            &va_block->read_duplicated_pages,
                                            resident_mask);
@@ -744,7 +781,7 @@ static NV_STATUS va_block_unset_read_duplication_locked(uvm_va_block_t *va_block
     }
 
     // 2- Re-establish SetAccessedBy mappings
-    for_each_id_in_mask(processor_id, &accessed_by) {
+    for_each_id_in_mask(processor_id, &policy->accessed_by) {
         status = uvm_va_block_set_accessed_by_locked(va_block,
                                                      va_block_context,
                                                      processor_id,
