@@ -1,5 +1,5 @@
 /*
- * SPDX-FileCopyrightText: Copyright (c) 1993-2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
+ * SPDX-FileCopyrightText: Copyright (c) 1993-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
  * SPDX-License-Identifier: MIT
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
@@ -30,6 +30,8 @@
 *                                                                          *
 \***************************************************************************/
 
+#define NVOC_KERNEL_SM_DEBUGGER_SESSION_H_PRIVATE_ACCESS_ALLOWED
+
 #include "kernel/rmapi/control.h"
 #include "kernel/rmapi/rmapi.h"
 #include "kernel/os/os.h"
@@ -58,14 +60,14 @@
 // for the caller to explicitly pass in the handle corresponding to the VaSpaceApi:
 //
 static NV_STATUS
-_nv83deCtrlCmdFetchVAS(NvU32 hClient, NvU32 hChannel, OBJVASPACE **ppVASpace)
+_nv83deCtrlCmdFetchVAS(RsClient *pClient, NvU32 hChannel, OBJVASPACE **ppVASpace)
 {
     KernelChannel *pKernelChannel = NULL;
 
     NV_ASSERT_OR_RETURN(ppVASpace != NULL, NV_ERR_INVALID_ARGUMENT);
 
     // Fetch the corresponding Channel object from our handle
-    NV_ASSERT_OK_OR_RETURN(CliGetKernelChannel(hClient, hChannel, &pKernelChannel));
+    NV_ASSERT_OK_OR_RETURN(CliGetKernelChannel(pClient, hChannel, &pKernelChannel));
     NV_ASSERT_OR_RETURN(pKernelChannel != NULL, NV_ERR_INVALID_ARGUMENT);
 
     *ppVASpace = pKernelChannel->pVAS;
@@ -148,7 +150,7 @@ _nv8deCtrlCmdReadWriteSurface
 )
 {
     OBJGPU *pGpu = GPU_RES_GET_GPU(pKernelSMDebuggerSession);
-    NvHandle hClient = RES_GET_CLIENT_HANDLE(pKernelSMDebuggerSession);
+    RsClient *pClient = RES_GET_CLIENT(pKernelSMDebuggerSession);
     OBJVASPACE *pVASpace = NULL;
     NvU32 count = pParams->count;
     NvU32 i;
@@ -161,7 +163,7 @@ _nv8deCtrlCmdReadWriteSurface
 
     // Attempt to retrieve the VAS pointer
     NV_ASSERT_OK_OR_RETURN(
-        _nv83deCtrlCmdFetchVAS(hClient, pKernelSMDebuggerSession->hChannel, &pVASpace));
+        _nv83deCtrlCmdFetchVAS(pClient, pKernelSMDebuggerSession->hChannel, &pVASpace));
 
     // Validate VA range and fail if invalid
     NV_ASSERT_OK_OR_RETURN(
@@ -169,7 +171,6 @@ _nv8deCtrlCmdReadWriteSurface
 
     for (i = 0; i < count; i++)
     {
-        NvU8 *pBase;
         MEMORY_DESCRIPTOR *pMemDesc = NULL;
         NvU64 virtAddr = pParams->opsBuffer[i].gpuVA;
         NvP64 bufPtr = pParams->opsBuffer[i].pCpuVA;
@@ -187,12 +188,15 @@ _nv8deCtrlCmdReadWriteSurface
         end4kPage = (NvOffset_LO32(virtAddr + bufSize - 1) >> 12) & 0x1FFFF;
 
         curSize = RM_PAGE_SIZE - pageStartOffset;
-        virtAddr &= ~RM_PAGE_MASK;
+        virtAddr -= pageStartOffset;
 
         for (cur4kPage = start4kPage; cur4kPage <= end4kPage; ++cur4kPage)
         {
-            MMU_TRACE_PARAM mmuParams = {0};
-            MMU_TRACE_ARG traceArg    = {0};
+            MMU_TRACE_PARAM mmuParams     = {0};
+            MMU_TRACE_ARG traceArg        = {0};
+            TRANSFER_SURFACE surf         = {0};
+            MemoryManager *pMemoryManager = GPU_GET_MEMORY_MANAGER(pGpu);
+            NvU8 *pKernBuffer = NULL;
 
             mmuParams.mode    = MMU_TRACE_MODE_TRANSLATE;
             mmuParams.va      = virtAddr;
@@ -207,17 +211,23 @@ _nv8deCtrlCmdReadWriteSurface
                 curSize = bufSize;
             }
 
+            pKernBuffer = portMemAllocNonPaged(curSize);
+            if (pKernBuffer == NULL)
+            {
+                return NV_ERR_INSUFFICIENT_RESOURCES;
+            }
+
             if (traceArg.aperture == ADDR_SYSMEM)
             {
                 NvP64 physAddr = NV_PTR_TO_NvP64(traceArg.pa);
-                NvU64 limit = (NvU64)(curSize - 1);
+                NvU64 limit = RM_PAGE_SIZE - 1;
 
                 NvU32 os02Flags = DRF_DEF(OS02, _FLAGS, _LOCATION,      _PCI)           |
                                   DRF_DEF(OS02, _FLAGS, _MAPPING,       _NO_MAP)        |
                                   DRF_DEF(OS02, _FLAGS, _PHYSICALITY,   _CONTIGUOUS)    |
                                   DRF_DEF(OS02, _FLAGS, _COHERENCY,     _CACHED);
 
-                NV_CHECK_OK_OR_RETURN(LEVEL_ERROR,
+                NV_ASSERT_OK_OR_ELSE(status,
                     osCreateMemFromOsDescriptor(pGpu,
                                                 physAddr,
                                                 pKernelSMDebuggerSession->hInternalClient,
@@ -225,39 +235,55 @@ _nv8deCtrlCmdReadWriteSurface
                                                 &limit,
                                                 &pMemDesc,
                                                 NVOS32_DESCRIPTOR_TYPE_OS_PHYS_ADDR,
-                                                RS_PRIV_LEVEL_KERNEL));
+                                                RS_PRIV_LEVEL_KERNEL),
+                    portMemFree(pKernBuffer); return status; );
             }
             else if (traceArg.aperture == ADDR_FBMEM)
             {
-                NV_ASSERT_OK_OR_RETURN(memdescCreate(&pMemDesc, pGpu, curSize, 0, NV_TRUE,
-                                                     traceArg.aperture, NV_MEMORY_UNCACHED, MEMDESC_FLAGS_NONE));
-                memdescDescribe(pMemDesc, traceArg.aperture, traceArg.pa, curSize);
+                NV_ASSERT_OK_OR_ELSE(status,
+                    memdescCreate(&pMemDesc, pGpu, RM_PAGE_SIZE, 0, NV_TRUE,
+                                  traceArg.aperture, NV_MEMORY_UNCACHED,
+                                  MEMDESC_FLAGS_NONE),
+                    portMemFree(pKernBuffer); return status; );
+                memdescDescribe(pMemDesc, traceArg.aperture, traceArg.pa, RM_PAGE_SIZE);
             }
 
-            pBase = kbusMapRmAperture_HAL(pGpu, pMemDesc);
-            NV_ASSERT_OR_ELSE(
-                pBase != NULL,
-                memdescDestroy(pMemDesc);
-                return NV_ERR_INVALID_ARGUMENT; );
+            surf.pMemDesc = pMemDesc;
+            surf.offset = pageStartOffset;
 
             if (bWrite)
             {
                 NV_ASSERT_OK_OR_CAPTURE_FIRST_ERROR(status,
-                    portMemExCopyFromUser(bufPtr, pBase + pageStartOffset, curSize));
+                    portMemExCopyFromUser(bufPtr, pKernBuffer, curSize));
+
+                // Write out the buffer to memory
+                if (status == NV_OK)
+                {
+                    NV_ASSERT_OK_OR_CAPTURE_FIRST_ERROR(status,
+                        memmgrMemWrite(pMemoryManager, &surf, pKernBuffer, curSize,
+                                       TRANSFER_FLAGS_DEFER_FLUSH));
+                }
             }
             else
             {
+                // Read from memory
                 NV_ASSERT_OK_OR_CAPTURE_FIRST_ERROR(status,
-                    portMemExCopyToUser(pBase + pageStartOffset, bufPtr, curSize));
+                    memmgrMemRead(pMemoryManager, &surf, pKernBuffer, curSize,
+                                  TRANSFER_FLAGS_DEFER_FLUSH));
+
+                if (status == NV_OK)
+                {
+                    NV_ASSERT_OK_OR_CAPTURE_FIRST_ERROR(status,
+                        portMemExCopyToUser(pKernBuffer, bufPtr, curSize));
+                }
             }
 
-            kbusUnmapRmAperture_HAL(pGpu, pMemDesc, &pBase, NV_FALSE);
+            portMemFree(pKernBuffer);
             memdescDestroy(pMemDesc);
 
             if (status != NV_OK)
                 return status;
 
-            pBase = NULL;
             pageStartOffset = 0;
             bufPtr = NvP64_PLUS_OFFSET(bufPtr,curSize);
             bufSize -= curSize;
@@ -297,7 +323,7 @@ ksmdbgssnCtrlCmdGetMappings_IMPL
 )
 {
     OBJGPU *pGpu = GPU_RES_GET_GPU(pKernelSMDebuggerSession);
-    NvHandle hClient = RES_GET_CLIENT_HANDLE(pKernelSMDebuggerSession);
+    RsClient *pClient = RES_GET_CLIENT(pKernelSMDebuggerSession);
     OBJVASPACE *pVASpace = NULL;
     MMU_TRACE_ARG traceArg = {0};
     MMU_TRACE_PARAM mmuParams = {0};
@@ -306,7 +332,7 @@ ksmdbgssnCtrlCmdGetMappings_IMPL
 
     // Attempt to retrieve the VAS pointer
     NV_ASSERT_OK_OR_RETURN(
-        _nv83deCtrlCmdFetchVAS(hClient, pKernelSMDebuggerSession->hChannel, &pVASpace));
+        _nv83deCtrlCmdFetchVAS(pClient, pKernelSMDebuggerSession->hChannel, &pVASpace));
 
     traceArg.pMapParams = pParams;
 
@@ -543,6 +569,13 @@ _nv83deCtrlCmdDebugAccessMemory
         flags = FLD_SET_DRF(OS33, _FLAGS, _MAPPING, _DIRECT, flags);
     }
 
+    // Allow the mapping to happen successfully on HCC devtools mode
+    if (kbusIsBarAccessBlocked(GPU_GET_KERNEL_BUS(pTargetGpu)) &&
+        gpuIsCCDevToolsModeEnabled(pTargetGpu))
+    {
+        flags = FLD_SET_DRF(OS33, _FLAGS, _ALLOW_MAPPING_ON_HCC, _YES, flags);
+    }
+
     // Map memory into the internal smdbg client
     rmStatus = _nv83deMapMemoryIntoGrdbgClient(pTargetGpu,
                                                pKernelSMDebuggerSession,
@@ -682,7 +715,6 @@ NV_STATUS ksmdbgssnCtrlCmdDebugExecRegOps_IMPL
 )
 {
     OBJGPU *pGpu = GPU_RES_GET_GPU(pKernelSMDebuggerSession);
-    NV_STATUS status = NV_OK;
     NvBool isClientGspPlugin = NV_FALSE;
 
     NV_CHECK_OR_RETURN(LEVEL_ERROR,
@@ -696,17 +728,16 @@ NV_STATUS ksmdbgssnCtrlCmdDebugExecRegOps_IMPL
 
     if (IS_GSP_CLIENT(pGpu))
     {
-        CALL_CONTEXT *pCallContext = resservGetTlsCallContext();
+        CALL_CONTEXT *pCallContext  = resservGetTlsCallContext();
         RmCtrlParams *pRmCtrlParams = pCallContext->pControlParams;
+        RM_API       *pRmApi        = GPU_GET_PHYSICAL_RMAPI(pGpu);
 
-        NV_RM_RPC_CONTROL(pGpu,
-                          pRmCtrlParams->hClient,
-                          pRmCtrlParams->hObject,
-                          pRmCtrlParams->cmd,
-                          pRmCtrlParams->pParams,
-                          pRmCtrlParams->paramsSize,
-                          status);
-        return status;
+        return pRmApi->Control(pRmApi,
+                               pRmCtrlParams->hClient,
+                               pRmCtrlParams->hObject,
+                               pRmCtrlParams->cmd,
+                               pRmCtrlParams->pParams,
+                               pRmCtrlParams->paramsSize);
     }
 
     return NV_ERR_NOT_SUPPORTED;
@@ -840,7 +871,7 @@ ksmdbgssnCtrlCmdDebugReadAllSmErrorStates_IMPL
         {
             NV_ASSERT_OK(
                 kgrctxLookupMmuFault(pGpu,
-                                     pKernelSMDebuggerSession->pObject->pKernelGraphicsContext,
+                                     kgrobjGetKernelGraphicsContext(pGpu, pKernelSMDebuggerSession->pObject),
                                      &pParams->mmuFault));
         }
 
@@ -884,7 +915,7 @@ ksmdbgssnCtrlCmdDebugClearAllSmErrorStates_IMPL
         if (IS_VIRTUAL_WITH_SRIOV(pGpu))
         {
             NV_ASSERT_OK(
-                kgrctxClearMmuFault(pGpu, pKernelSMDebuggerSession->pObject->pKernelGraphicsContext));
+                kgrctxClearMmuFault(pGpu, kgrobjGetKernelGraphicsContext(pGpu, pKernelSMDebuggerSession->pObject)));
         }
 
         return rmStatus;
@@ -929,7 +960,7 @@ ksmdbgssnCtrlCmdDebugReadMMUFaultInfo_IMPL
 
     NV_ASSERT_OK_OR_RETURN(
         kgrctxLookupMmuFaultInfo(pGpu,
-                                 pKernelSMDebuggerSession->pObject->pKernelGraphicsContext,
+                                 kgrobjGetKernelGraphicsContext(pGpu, pKernelSMDebuggerSession->pObject),
                                  pParams));
 
     return NV_OK;
